# Copyright (c) 2010-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for loading-related functionality in scene.py."""

from unittest import mock

import pytest
import xarray as xr
from dask import array as da

from satpy import Scene
from satpy.tests.utils import make_cid, make_dataid, make_dsq, spy_decorator

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - include_test_etc


@pytest.mark.usefixtures("include_test_etc")
class TestSceneAllAvailableDatasets:
    """Test the Scene's handling of various dependencies."""

    def test_all_datasets_no_readers(self):
        """Test all datasets with no reader."""
        scene = Scene()
        pytest.raises(KeyError, scene.all_dataset_ids, reader_name="fake")
        id_list = scene.all_dataset_ids()
        assert id_list == []
        # no sensors are loaded so we shouldn't get any comps either
        id_list = scene.all_dataset_ids(composites=True)
        assert id_list == []

    def test_all_dataset_names_no_readers(self):
        """Test all dataset names with no reader."""
        scene = Scene()
        pytest.raises(KeyError, scene.all_dataset_names, reader_name="fake")
        name_list = scene.all_dataset_names()
        assert name_list == []
        # no sensors are loaded so we shouldn't get any comps either
        name_list = scene.all_dataset_names(composites=True)
        assert name_list == []

    def test_available_dataset_no_readers(self):
        """Test the available datasets without a reader."""
        scene = Scene()
        pytest.raises(
            KeyError, scene.available_dataset_ids, reader_name="fake")
        name_list = scene.available_dataset_ids()
        assert name_list == []
        # no sensors are loaded so we shouldn't get any comps either
        name_list = scene.available_dataset_ids(composites=True)
        assert name_list == []

    def test_available_dataset_names_no_readers(self):
        """Test the available dataset names without a reader."""
        scene = Scene()
        pytest.raises(
            KeyError, scene.available_dataset_names, reader_name="fake")
        name_list = scene.available_dataset_names()
        assert name_list == []
        # no sensors are loaded so we shouldn't get any comps either
        name_list = scene.available_dataset_names(composites=True)
        assert name_list == []

    def test_all_datasets_one_reader(self):
        """Test all datasets for one reader."""
        scene = Scene(filenames=["fake1_1.txt"],
                      reader="fake1")
        id_list = scene.all_dataset_ids()
        # 20 data products + 6 lon/lat products
        num_reader_ds = 21 + 6
        assert len(id_list) == num_reader_ds
        id_list = scene.all_dataset_ids(composites=True)
        assert len(id_list) == num_reader_ds + 33

    def test_all_datasets_multiple_reader(self):
        """Test all datasets for multiple readers."""
        scene = Scene(filenames={"fake1_1ds": ["fake1_1ds_1.txt"],
                                 "fake2_1ds": ["fake2_1ds_1.txt"]})
        id_list = scene.all_dataset_ids()
        assert len(id_list) == 2
        id_list = scene.all_dataset_ids(composites=True)
        # ds1 and ds2 => 2
        # composites that use these two datasets => 11
        assert len(id_list) == 2 + 11

    def test_available_datasets_one_reader(self):
        """Test the available datasets for one reader."""
        scene = Scene(filenames=["fake1_1ds_1.txt"],
                      reader="fake1_1ds")
        id_list = scene.available_dataset_ids()
        assert len(id_list) == 1
        id_list = scene.available_dataset_ids(composites=True)
        # ds1, comp1, comp14, comp16, static_image, comp26
        assert len(id_list) == 6

    def test_available_composite_ids_missing_available(self):
        """Test available_composite_ids when a composites dep is missing."""
        scene = Scene(filenames=["fake1_1ds_1.txt"],
                      reader="fake1_1ds")
        assert "comp2" not in scene.available_composite_names()

    def test_available_composites_known_versus_all(self):
        """Test available_composite_ids when some datasets aren't available."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1",
                      reader_kwargs={"not_available": ["ds2", "ds3"]})
        all_comps = scene.all_composite_names()
        avail_comps = scene.available_composite_names()
        # there should always be more known composites than available composites
        assert len(all_comps) > len(avail_comps)
        for not_avail_comp in ("comp2", "comp3"):
            assert not_avail_comp in all_comps
            assert not_avail_comp not in avail_comps

    def test_available_comps_no_deps(self):
        """Test Scene available composites when composites don't have a dependency."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        all_comp_ids = scene.available_composite_ids()
        assert make_cid(name="static_image") in all_comp_ids
        available_comp_ids = scene.available_composite_ids()
        assert make_cid(name="static_image") in available_comp_ids

    def test_available_when_sensor_none_in_preloaded_dataarrays(self):
        """Test Scene available composites when existing loaded arrays have sensor set to None.

        Some readers or composites (ex. static images) don't have a sensor and
        developers choose to set it to `None`. This test makes sure this
        doesn't break available composite IDs.

        """
        scene = _scene_with_data_array_none_sensor()
        available_comp_ids = scene.available_composite_ids()
        assert make_cid(name="static_image") in available_comp_ids


@pytest.mark.usefixtures("include_test_etc")
class TestBadLoading:
    """Test the Scene object's `.load` method with bad inputs."""

    def test_load_str(self):
        """Test passing a string to Scene.load."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        pytest.raises(TypeError, scene.load, "ds1")

    def test_load_no_exist(self):
        """Test loading a dataset that doesn't exist."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        pytest.raises(KeyError, scene.load, ["im_a_dataset_that_doesnt_exist"])


@pytest.mark.usefixtures("include_test_etc")
class TestLoadingReaderDatasets:
    """Test the Scene object's `.load` method for datasets coming from a reader."""

    def test_load_no_exist2(self):
        """Test loading a dataset that doesn't exist then another load."""
        from satpy.readers.core.yaml_reader import FileYAMLReader
        load_mock = spy_decorator(FileYAMLReader.load)
        with mock.patch.object(FileYAMLReader, "load", load_mock):
            lmock = load_mock.mock
            scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
            scene.load(["ds9_fail_load"])
            loaded_ids = list(scene._datasets.keys())
            assert len(loaded_ids) == 0
            lmock.assert_called_once_with(
                {make_dataid(name="ds9_fail_load", wavelength=(1.0, 1.1, 1.2))})

            scene.load(["ds1"])
            loaded_ids = list(scene._datasets.keys())
            assert lmock.call_count == 2
            # most recent call should have only been ds1
            lmock.assert_called_with({
                make_dataid(name="ds1", resolution=250, calibration="reflectance", modifiers=tuple()),
            })
            assert len(loaded_ids) == 1

    def test_load_ds1_no_comps(self):
        """Test loading one dataset with no loaded compositors."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["ds1"])
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 1
        assert loaded_ids[0] == make_dataid(name="ds1", resolution=250, calibration="reflectance", modifiers=tuple())

    def test_load_ds1_load_twice(self):
        """Test loading one dataset with no loaded compositors."""
        from satpy.readers.core.yaml_reader import FileYAMLReader
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["ds1"])
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 1
        assert loaded_ids[0] == make_dataid(name="ds1", resolution=250, calibration="reflectance", modifiers=tuple())

        load_mock = spy_decorator(FileYAMLReader.load)
        with mock.patch.object(FileYAMLReader, "load", load_mock):
            lmock = load_mock.mock
            scene.load(["ds1"])
            loaded_ids = list(scene._datasets.keys())
            assert len(loaded_ids) == 1
            assert loaded_ids[0] == make_dataid(name="ds1",
                                                resolution=250,
                                                calibration="reflectance",
                                                modifiers=tuple())
            assert not lmock.called, ("Reader.load was called again when "
                                      "loading something that's already "
                                      "loaded")

    def test_load_ds1_unknown_modifier(self):
        """Test loading one dataset with no loaded compositors."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        pytest.raises(KeyError, scene.load,
                      [make_dataid(name="ds1", modifiers=("_fake_bad_mod_",))])

    def test_load_ds4_cal(self):
        """Test loading a dataset that has two calibration variations."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["ds4"])
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 1
        assert loaded_ids[0]["calibration"] == "reflectance"

    @pytest.mark.parametrize(
        ("input_filenames", "load_kwargs", "exp_resolution"),
        [
            (["fake1_1.txt", "fake1_highres_1.txt"], {}, 250),
            (["fake1_1.txt"], {"resolution": [500, 1000]}, 500),
            (["fake1_1.txt"], {"modifiers": tuple()}, 500),
            (["fake1_1.txt"], {}, 500),
        ]
    )
    def test_load_ds5_variations(self, input_filenames, load_kwargs, exp_resolution):
        """Test loading a dataset has multiple resolutions available."""
        scene = Scene(filenames=input_filenames, reader="fake1")
        scene.load(["ds5"], **load_kwargs)
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 1
        assert loaded_ids[0]["name"] == "ds5"
        assert loaded_ids[0]["resolution"] == exp_resolution

    def test_load_ds5_multiple_resolution_loads(self):
        """Test loading a dataset with multiple resolutions available as separate loads."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["ds5"], resolution=1000)
        scene.load(["ds5"], resolution=500)
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 2
        assert loaded_ids[0]["name"] == "ds5"
        assert loaded_ids[0]["resolution"] == 500
        assert loaded_ids[1]["name"] == "ds5"
        assert loaded_ids[1]["resolution"] == 1000

    def test_load_ds6_wl(self):
        """Test loading a dataset by wavelength."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load([0.22])
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 1
        assert loaded_ids[0]["name"] == "ds6"

    def test_load_ds9_fail_load(self):
        """Test loading a dataset that will fail during load."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["ds9_fail_load"])
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 0


@pytest.mark.usefixtures("include_test_etc")
class TestLoadingComposites:
    """Test the Scene object's `.load` method for composites."""

    @pytest.mark.parametrize(
        ("comp_name", "exp_id_or_name"),
        [
            pytest.param("comp1", make_cid(name="comp1"), id="composite with one required reader prereq"),
            pytest.param("comp4", make_cid(name="comp4"), id="composite with a required composite prereq"),
            pytest.param("comp5", make_cid(name="comp5"), id="composite with an optional reader prereq"),
            pytest.param("comp6", make_cid(name="comp6"), id="composite with an optional composite prereq"),
            pytest.param("comp9", make_cid(name="comp9"), id="composite with an unknown optional prereq"),
            pytest.param("comp10", make_cid(name="comp10"), id="composite with a modified required prereq"),
            pytest.param("comp11", make_cid(name="comp11"), id="composite with required prereqs as wavelength"),
            pytest.param("comp12", make_cid(name="comp12"),
                         id="composite with required prereqs as modified wavelengths"),
            pytest.param("comp13", make_cid(name="comp13"), id="composite with modified res-changed prereq"),
            pytest.param("comp14", make_cid(name="comp14", resolution=555),
                         id="composite that changes DataID resolution"),
            pytest.param("comp16", make_cid(name="comp16"), id="composite with unloadable optional prereq"),
            pytest.param("comp20", make_cid(name="comp20"), id="composite with prereq with modifier with opt prereq"),
            pytest.param("comp21", make_cid(name="comp21"),
                         id="composite with prereq with modifier with unloadable opt prereq"),
            pytest.param("comp22", make_cid(name="comp22"),
                         id="composite with prereq with modifier with only opt prereqs"),
            pytest.param("ahi_green", make_cid(name="ahi_green"), id="ahi_green composite"),
        ]
    )
    def test_single_composite_loading(self, comp_name, exp_id_or_name):
        """Test that certain composites can be loaded individually."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load([comp_name])
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 1
        if isinstance(exp_id_or_name, str):
            assert loaded_ids[0]["name"] == exp_id_or_name
        else:
            assert loaded_ids[0] == exp_id_or_name

    def test_load_multiple_resolutions(self):
        """Test loading a dataset has multiple resolutions available with different resolutions."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        comp25 = make_cid(name="comp25", resolution=1000)
        scene[comp25] = xr.DataArray([], attrs={"name": "comp25", "resolution": 1000})
        scene.load(["comp25"], resolution=500)

        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 2
        assert loaded_ids[0]["name"] == "comp25"
        assert loaded_ids[0]["resolution"] == 500
        assert loaded_ids[1]["name"] == "comp25"
        assert loaded_ids[1]["resolution"] == 1000

    def test_load_same_subcomposite(self):
        """Test loading a composite and one of it's subcomposites at the same time."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["comp24", "comp25"], resolution=500)
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 2
        assert loaded_ids[0]["name"] == "comp24"
        assert loaded_ids[0]["resolution"] == 500
        assert loaded_ids[1]["name"] == "comp25"
        assert loaded_ids[1]["resolution"] == 500

    def test_load_comp8(self):
        """Test loading a composite that has a non-existent prereq."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        pytest.raises(KeyError, scene.load, ["comp8"])

    def test_load_comp15(self):
        """Test loading a composite whose prerequisites can't be loaded.

        Note that the prereq exists in the reader, but fails in loading.

        """
        # it is fine that an optional prereq doesn't exist
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["comp15"])
        loaded_ids = list(scene._datasets.keys())
        assert not loaded_ids

    def test_load_comp17(self):
        """Test loading a composite that depends on a composite that won't load."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["comp17"])
        loaded_ids = list(scene._datasets.keys())
        assert not loaded_ids

    def test_load_comp18(self):
        """Test loading a composite that depends on an incompatible area modified dataset."""
        # it is fine that an optional prereq doesn't exist
        scene = Scene(filenames=["fake1_1.txt", "fake1_highres_1.txt"], reader="fake1")
        scene.load(["comp18"])
        loaded_ids = list(scene._datasets.keys())
        # depends on:
        #   ds3
        #   ds4 (mod1, mod3)
        #   ds5 (mod1, incomp_areas)
        # We should end up with ds3, ds4 (mod1, mod3), ds5 (mod1), and ds1
        # for the incomp_areas modifier
        assert len(loaded_ids) == 4  # the 1 dependencies
        assert "ds3" in scene._datasets
        assert make_dataid(name="ds4", calibration="reflectance",
                           modifiers=("mod1", "mod3")) in scene._datasets
        assert make_dataid(name="ds5", resolution=250,
                           modifiers=("mod1",)) in scene._datasets

    def test_load_comp18_2(self):
        """Test loading a composite that depends on an incompatible area modified dataset.

        Specifically a modified dataset where the modifier has optional
        dependencies.

        """
        # it is fine that an optional prereq doesn't exist
        scene = Scene(filenames=["fake1_1.txt", "fake1_highres_1.txt"], reader="fake1")
        scene.load(["comp18_2"])
        loaded_ids = list(scene._datasets.keys())
        # depends on:
        #   ds3
        #   ds4 (mod1, mod3)
        #   ds5 (mod1, incomp_areas_opt)
        # We should end up with ds3, ds4 (mod1, mod3), ds5 (mod1), and ds1
        # and ds2 for the incomp_areas_opt modifier
        assert len(loaded_ids) == 5  # the 1 dependencies
        assert "ds3" in scene._datasets
        assert "ds2" in scene._datasets
        assert make_dataid(name="ds4", calibration="reflectance",
                           modifiers=("mod1", "mod3")) in scene._datasets
        assert make_dataid(name="ds5", resolution=250,
                           modifiers=("mod1",)) in scene._datasets

    def test_load_comp19(self):
        """Test loading a composite that shares a dep with a dependency.

        More importantly test that loading a dependency that depends on
        the same dependency as this composite (a sibling dependency) and
        that sibling dependency includes a modifier. This test makes sure
        that the Node in the dependency tree is the exact same node.

        """
        # Check dependency tree nodes
        # initialize the dep tree without loading the data
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene._update_dependency_tree({"comp19"}, None)

        this_node = scene._dependency_tree["comp19"]
        shared_dep_id = make_dataid(name="ds5", modifiers=("res_change",))
        shared_dep_expected_node = scene._dependency_tree[shared_dep_id]
        # get the node for the first dep in the prereqs list of the
        # comp13 node
        shared_dep_node = scene._dependency_tree["comp13"].data[1][0]
        shared_dep_node2 = this_node.data[1][0]
        assert shared_dep_expected_node is shared_dep_node
        assert shared_dep_expected_node is shared_dep_node2

        scene.load(["comp19"])

        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 1
        assert loaded_ids[0] == make_cid(name="comp19")

    def test_load_multiple_comps(self):
        """Test loading multiple composites."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["comp1", "comp2", "comp3", "comp4", "comp5", "comp6",
                    "comp7", "comp9", "comp10"])
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 9

    def test_load_multiple_comps_separate(self):
        """Test loading multiple composites, one at a time."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["comp10"])
        scene.load(["comp9"])
        scene.load(["comp7"])
        scene.load(["comp6"])
        scene.load(["comp5"])
        scene.load(["comp4"])
        scene.load(["comp3"])
        scene.load(["comp2"])
        scene.load(["comp1"])
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 9

    def test_load_modified(self):
        """Test loading a modified dataset."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load([make_dsq(name="ds1", modifiers=("mod1", "mod2"))])
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 1
        assert loaded_ids[0]["modifiers"] == ("mod1", "mod2")

    def test_load_modified_with_load_kwarg(self):
        """Test loading a modified dataset using the ``Scene.load`` keyword argument."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["ds1"], modifiers=("mod1", "mod2"))
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 1
        assert loaded_ids[0]["modifiers"] == ("mod1", "mod2")

    def test_load_multiple_modified(self):
        """Test loading multiple modified datasets."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load([
            make_dataid(name="ds1", modifiers=("mod1", "mod2")),
            make_dataid(name="ds2", modifiers=("mod2", "mod1")),
        ])
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 2
        for i in loaded_ids:
            if i["name"] == "ds1":
                assert i["modifiers"] == ("mod1", "mod2")
            else:
                assert i["name"] == "ds2"
                assert i["modifiers"] == ("mod2", "mod1")

    def test_load_dataset_after_composite(self):
        """Test load composite followed by other datasets."""
        from satpy.readers.core.yaml_reader import FileYAMLReader
        from satpy.tests.utils import FakeCompositor
        load_mock = spy_decorator(FileYAMLReader.load)
        comp_mock = spy_decorator(FakeCompositor.__call__)
        with mock.patch.object(FileYAMLReader, "load", load_mock), \
                mock.patch.object(FakeCompositor, "__call__", comp_mock):
            lmock = load_mock.mock
            scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
            scene.load(["comp3"])
            assert lmock.call_count == 1
            scene.load(["ds1"])
            assert lmock.call_count == 2
            scene.load(["ds1"])
            # we should only load from the file twice
            assert lmock.call_count == 2
            # we should only generate the composite once
            assert comp_mock.mock.call_count == 1
            loaded_ids = list(scene._datasets.keys())
            assert len(loaded_ids) == 2

    def test_load_dataset_after_composite2(self):
        """Test load complex composite followed by other datasets."""
        from satpy.readers.core.yaml_reader import FileYAMLReader
        from satpy.tests.utils import FakeCompositor, FakeModifier
        load_mock = spy_decorator(FileYAMLReader.load)
        comp_mock = spy_decorator(FakeCompositor.__call__)
        mod_mock = spy_decorator(FakeModifier.__call__)
        with mock.patch.object(FileYAMLReader, "load", load_mock), \
             mock.patch.object(FakeCompositor, "__call__", comp_mock), \
             mock.patch.object(FakeModifier, "__call__", mod_mock):
            lmock = load_mock.mock
            scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
            scene.load(["comp10"])
            assert lmock.call_count == 1
            loaded_ids = list(scene._datasets.keys())
            assert len(loaded_ids) == 1
            with mock.patch.object(scene, "_generate_composites_nodes_from_loaded_datasets",
                                   wraps=scene._generate_composites_nodes_from_loaded_datasets) as m:
                scene.load(["ds1"])
                assert lmock.call_count == 2
                loaded_ids = list(scene._datasets.keys())
                assert len(loaded_ids) == 2
                # this is the unmodified ds1
                assert make_dataid(
                    name="ds1", resolution=250, calibration="reflectance", modifiers=tuple()
                ) in loaded_ids
                # m.assert_called_once_with(set([scene._dependency_tree['ds1']]))
                m.assert_called_once_with(set())
            with mock.patch.object(scene, "_generate_composites_nodes_from_loaded_datasets",
                                   wraps=scene._generate_composites_nodes_from_loaded_datasets) as m:
                scene.load(["ds1"])
                assert lmock.call_count == 2
                loaded_ids = list(scene._datasets.keys())
                assert len(loaded_ids) == 2
                # this is the unmodified ds1
                assert make_dataid(
                    name="ds1", resolution=250, calibration="reflectance", modifiers=tuple()
                ) in loaded_ids
                m.assert_called_once_with(set())
            # we should only generate the comp10 composite once but comp2 was also generated
            assert comp_mock.mock.call_count == 1 + 1
            # Create the modded ds1 at comp10, then load the umodified version
            # again
            assert mod_mock.mock.call_count == 1
            loaded_ids = list(scene._datasets.keys())
            assert len(loaded_ids) == 2

    def test_no_generate_comp10(self):
        """Test generating a composite after loading."""
        # it is fine that an optional prereq doesn't exist
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["comp10"], generate=False)
        assert any(ds_id["name"] == "comp10" for ds_id in scene._wishlist)
        assert "comp10" not in scene._datasets
        # two dependencies should have been loaded
        assert len(scene._datasets) == 2
        assert len(scene.missing_datasets) == 1

        scene._generate_composites_from_loaded_datasets()
        assert any(ds_id["name"] == "comp10" for ds_id in scene._wishlist)
        assert "comp10" in scene._datasets
        assert not scene.missing_datasets

    def test_modified_with_wl_dep(self):
        """Test modifying a dataset with a modifier with modified deps.

        More importantly test that loading the modifiers dependency at the
        same time as the original modified dataset that the dependency tree
        nodes are unique and that DataIDs.

        """
        from satpy.dataset.dataid import WavelengthRange

        # Check dependency tree nodes
        # initialize the dep tree without loading the data
        ds1_mod_id = make_dsq(name="ds1", modifiers=("mod_wl",))
        ds3_mod_id = make_dsq(name="ds3", modifiers=("mod_wl",))

        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene._update_dependency_tree({ds1_mod_id, ds3_mod_id}, None)

        ds1_mod_node = scene._dependency_tree[ds1_mod_id]
        ds3_mod_node = scene._dependency_tree[ds3_mod_id]
        ds1_mod_dep_node = ds1_mod_node.data[1][1]
        ds3_mod_dep_node = ds3_mod_node.data[1][1]
        # mod_wl depends on the this node:
        ds6_modded_node = scene._dependency_tree[make_dataid(name="ds6", modifiers=("mod1",))]
        # this dep should be full qualified with name and wavelength
        assert ds6_modded_node.name["name"] is not None
        assert isinstance(ds6_modded_node.name["wavelength"], WavelengthRange)
        # the node should be shared between everything that uses it
        assert ds1_mod_dep_node is ds3_mod_dep_node
        assert ds1_mod_dep_node is ds6_modded_node

        # it is fine that an optional prereq doesn't exist
        scene.load([ds1_mod_id, ds3_mod_id])

        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 2
        assert ds1_mod_id in scene._datasets
        assert ds3_mod_id in scene._datasets

    def test_load_comp11_and_23(self):
        """Test loading two composites that depend on similar wavelengths."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        # mock the available comps/mods in the compositor loader
        avail_comps = scene.available_composite_ids()
        assert make_cid(name="comp11") in avail_comps
        assert make_cid(name="comp23") in avail_comps
        # it is fine that an optional prereq doesn't exist
        scene.load(["comp11", "comp23"])
        comp11_node = scene._dependency_tree["comp11"]
        comp23_node = scene._dependency_tree["comp23"]
        assert comp11_node.data[1][-1].name["name"] == "ds10"
        assert comp23_node.data[1][0].name["name"] == "ds8"
        loaded_ids = list(scene._datasets.keys())
        assert len(loaded_ids) == 2
        assert "comp11" in scene
        assert "comp23" in scene

    def test_load_too_many(self):
        """Test dependency tree if too many reader keys match."""
        scene = Scene(filenames=["fake3_1.txt"], reader="fake3")
        avail_comps = scene.available_composite_ids()
        # static image => 1
        assert len(avail_comps) == 1
        pytest.raises(KeyError, scene.load, [0.21])

    def test_load_when_sensor_none_in_preloaded_dataarrays(self):
        """Test Scene loading when existing loaded arrays have sensor set to None.

        Some readers or composites (ex. static images) don't have a sensor and
        developers choose to set it to `None`. This test makes sure this
        doesn't break loading.

        """
        scene = _scene_with_data_array_none_sensor()
        scene.load(["static_image"])
        assert "static_image" in scene
        assert "my_data" in scene


def _scene_with_data_array_none_sensor():
    scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
    scene["my_data"] = _data_array_none_sensor("my_data")
    return scene


def _data_array_none_sensor(name: str) -> xr.DataArray:
    """Create a DataArray with sensor set to ``None``."""
    return xr.DataArray(
        da.zeros((2, 2)),
        attrs={
            "name": name,
            "sensor": None,
        })



# -*- coding: utf-8 -*-
# Copyright (c) 2020,2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""HY-2B L2B Reader.

Distributed by Eumetsat in HDF5 format.
Also handle the HDF5 files from NSOAS, based on a file example.
"""

import datetime as dt

import numpy as np
import xarray as xr

from satpy.readers.core.hdf5 import HDF5FileHandler


class HY2SCATL2BH5FileHandler(HDF5FileHandler):
    """File handler for HY2 scat."""

    @property
    def start_time(self):
        """Time for first observation."""
        return dt.datetime.strptime(self["/attr/Range_Beginning_Time"],
                                    "%Y%m%dT%H:%M:%S")

    @property
    def end_time(self):
        """Time for final observation."""
        return dt.datetime.strptime(self["/attr/Range_Ending_Time"],
                                    "%Y%m%dT%H:%M:%S")

    @property
    def platform_name(self):
        """Get the Platform ShortName."""
        return self["/attr/Platform_ShortName"]

    def get_variable_metadata(self):
        """Get the variable metadata."""
        info = getattr(self, "attrs", {})
        info.update({
            "Equator_Crossing_Longitude": self["/attr/Equator_Crossing_Longitude"],
            "Equator_Crossing_Time": self["/attr/Equator_Crossing_Time"],
            "Input_L2A_Filename": self["/attr/Input_L2A_Filename"],
            "L2B_Actual_WVC_Rows": self["/attr/L2B_Actual_WVC_Rows"],
            "Orbit_Inclination": self["/attr/Orbit_Inclination"],
            "Orbit_Number": self["/attr/Orbit_Number"],
            "Output_L2B_Filename": self["/attr/Output_L2B_Filename"],
            "Production_Date_Time": self["/attr/Production_Date_Time"],
            "L2B_Expected_WVC_Rows": self["/attr/L2B_Expected_WVC_Rows"]
        })
        try:
            info.update({"L2B_Number_WVC_cells": self["/attr/L2B_Number_WVC_cells"]})
        except KeyError:
            info.update({"L2B_Expected_WVC_Cells": self["/attr/L2B_Expected_WVC_Cells"]})
        return info

    def get_metadata(self):
        """Get the metadata."""
        info = getattr(self, "attrs", {})
        info.update({
            "WVC_Size": self["/attr/WVC_Size"],
            "HDF_Version_Id": self["/attr/HDF_Version_Id"],
            "Instrument_ShorName": self["/attr/Instrument_ShorName"],
            "L2A_Inputdata_Version": self["/attr/L2A_Inputdata_Version"],
            "L2B_Algorithm_Descriptor": self["/attr/L2B_Algorithm_Descriptor"],
            "L2B_Data_Version": self["/attr/L2B_Data_Version"],
            "L2B_Processing_Type": self["/attr/L2B_Processing_Type"],
            "L2B_Processor_Name": self["/attr/L2B_Processor_Name"],
            "L2B_Processor_Version": self["/attr/L2B_Processor_Version"],
            "Long_Name": self["/attr/Long_Name"],
            "Platform_LongName": self["/attr/Platform_LongName"],
            "Platform_ShortName": self["/attr/Platform_ShortName"],
            "Platform_Type": self["/attr/Platform_Type"],
            "Producer_Agency": self["/attr/Producer_Agency"],
            "Producer_Institution": self["/attr/Producer_Institution"],
            "Rev_Orbit_Perio": self["/attr/Rev_Orbit_Period"],
            "Short_Name": self["/attr/Short_Name"],
            "Sigma0_Granularity": self["/attr/Sigma0_Granularity"],
        })
        return info

    def get_dataset(self, key, info):
        """Get the dataset."""
        dims = ["y", "x"]
        if self[key["name"]].ndim == 3:
            dims = ["y", "x", "selection"]
        data = self[key["name"]]
        if "valid range" in data.attrs:
            data.attrs.update({"valid_range": data.attrs.pop("valid range")})
        if key["name"] in "wvc_row_time":
            data = data.rename({data.dims[0]: "y"})
        else:
            dim_map = {curr_dim: new_dim for curr_dim, new_dim in zip(data.dims, dims)}
            data = data.rename(dim_map)
            data = self._mask_data(data)
            data = self._scale_data(data)

            if key["name"] in "wvc_lon":
                _attrs = data.attrs
                data = xr.where(data > 180, data - 360., data)
                data.attrs.update(_attrs)
        data.attrs.update(info)
        data.attrs.update(self.get_metadata())
        data.attrs.update(self.get_variable_metadata())
        if "Platform_ShortName" in data.attrs:
            data.attrs.update({"platform_name": data.attrs["Platform_ShortName"]})

        return data

    def _scale_data(self, data):
        return data * data.attrs["scale_factor"] + data.attrs["add_offset"]

    def _mask_data(self, data):
        _attrs = data.attrs
        valid_range = data.attrs["valid_range"]
        data = xr.where(data == data.attrs["fill_value"], np.nan, data)
        data = xr.where(data < valid_range[0], np.nan, data)
        data = xr.where(data > valid_range[1], np.nan, data)
        data.attrs.update(_attrs)
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the 'seadas_l2' reader."""

import numpy as np
import pytest
from pyresample.geometry import SwathDefinition
from pytest_lazy_fixtures import lf as lazy_fixture

from satpy import Scene, available_readers

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path_factory


@pytest.fixture(scope="module")
def seadas_l2_modis_chlor_a(tmp_path_factory):
    """Create MODIS SEADAS file."""
    filename = "a1.21322.1758.seadas.hdf"
    full_path = str(tmp_path_factory.mktemp("seadas_l2") / filename)
    return _create_seadas_chlor_a_hdf4_file(full_path, "Aqua", "MODISA")


@pytest.fixture(scope="module")
def seadas_l2_viirs_npp_chlor_a(tmp_path_factory):
    """Create VIIRS NPP SEADAS file."""
    filename = "SEADAS_npp_d20211118_t1728125_e1739327.hdf"
    full_path = str(tmp_path_factory.mktemp("seadas") / filename)
    return _create_seadas_chlor_a_hdf4_file(full_path, "NPP", "VIIRSN")


@pytest.fixture(scope="module")
def seadas_l2_viirs_j01_chlor_a(tmp_path_factory):
    """Create VIIRS JPSS-01 SEADAS file."""
    filename = "SEADAS_j01_d20211118_t1728125_e1739327.hdf"
    full_path = str(tmp_path_factory.mktemp("seadas") / filename)
    return _create_seadas_chlor_a_hdf4_file(full_path, "JPSS-1", "VIIRSJ1")


def _create_seadas_chlor_a_hdf4_file(full_path, mission, sensor):
    from pyhdf.SD import SD, SDC
    h = SD(full_path, SDC.WRITE | SDC.CREATE)
    setattr(h, "Sensor Name", sensor)
    h.Mission = mission
    setattr(h, "Start Time", "2021322175853191")
    setattr(h, "End Time", "2021322180551214")

    lon_info = {
        "type": SDC.FLOAT32,
        "data": np.zeros((5, 5), dtype=np.float32),
        "dim_labels": ["Number of Scan Lines", "Number of Pixel Control Points"],
        "attrs": {
            "long_name": "Longitude\x00",
            "standard_name": "longitude\x00",
            "units": "degrees_east\x00",
            "valid_range": (-180.0, 180.0),
        }
    }
    lat_info = {
        "type": SDC.FLOAT32,
        "data": np.zeros((5, 5), np.float32),
        "dim_labels": ["Number of Scan Lines", "Number of Pixel Control Points"],
        "attrs": {
            "long_name": "Latitude\x00",
            "standard_name": "latitude\x00",
            "units": "degrees_north\x00",
            "valid_range": (-90.0, 90.0),
        }
    }
    _add_variable_to_hdf4_file(h, "longitude", lon_info)
    _add_variable_to_hdf4_file(h, "latitude", lat_info)

    chlor_a_info = {
        "type": SDC.FLOAT32,
        "data": np.ones((5, 5), np.float32),
        "dim_labels": ["Number of Scan Lines", "Number of Pixel Control Points"],
        "attrs": {
            "long_name": "Chlorophyll Concentration, OCI Algorithm\x00",
            "units": "mg m^-3\x00",
            "standard_name": "mass_concentration_of_chlorophyll_in_sea_water\x00",
            "valid_range": (0.001, 100.0),
        }
    }
    _add_variable_to_hdf4_file(h, "chlor_a", chlor_a_info)

    l2_flags = np.zeros((5, 5), dtype=np.int32)
    l2_flags[2, 2] = -1
    l2_flags_info = {
        "type": SDC.INT32,
        "data": l2_flags,
        "dim_labels": ["Number of Scan Lines", "Number of Pixel Control Points"],
        "attrs": {},
    }
    _add_variable_to_hdf4_file(h, "l2_flags", l2_flags_info)
    return [full_path]


def _add_variable_to_hdf4_file(h, var_name, var_info):
    v = h.create(var_name, var_info["type"], var_info["data"].shape)
    v[:] = var_info["data"]
    for dim_count, dimension_name in enumerate(var_info["dim_labels"]):
        v.dim(dim_count).setname(dimension_name)
    if var_info.get("fill_value"):
        v.setfillvalue(var_info["fill_value"])
    for attr_key, attr_val in var_info["attrs"].items():
        setattr(v, attr_key, attr_val)


@pytest.fixture(scope="module")
def seadas_l2_modis_chlor_a_netcdf(tmp_path_factory):
    """Create MODIS SEADAS NetCDF file."""
    filename = "t1.21332.1758.seadas.nc"
    full_path = str(tmp_path_factory.mktemp("seadas_l2") / filename)
    return _create_seadas_chlor_a_netcdf_file(full_path, "Terra", "MODIS")


def _create_seadas_chlor_a_netcdf_file(full_path, mission, sensor):
    from netCDF4 import Dataset
    nc = Dataset(full_path, "w")
    nc.createDimension("number_of_lines", 5)
    nc.createDimension("pixels_per_line", 5)
    nc.instrument = sensor
    nc.platform = mission
    nc.time_coverage_start = "2021-11-18T17:58:53.191Z"
    nc.time_coverage_end = "2021-11-18T18:05:51.214Z"

    lon_info = {
        "data": np.zeros((5, 5), dtype=np.float32),
        "dim_labels": ("number_of_lines", "pixels_per_line"),
        "attrs": {
            "long_name": "Longitude",
            "standard_name": "longitude",
            "units": "degrees_east",
            "valid_min": -180.0,
            "valid_max": 180.0,
        }
    }
    lat_info = {
        "data": np.zeros((5, 5), np.float32),
        "dim_labels": ("number_of_lines", "pixels_per_line"),
        "attrs": {
            "long_name": "Latitude",
            "standard_name": "latitude",
            "units": "degrees_north",
            "valid_min": -90.0,
            "valid_max": 90.0,
        }
    }
    nav_group = nc.createGroup("navigation_data")
    _add_variable_to_netcdf_file(nav_group, "longitude", lon_info)
    _add_variable_to_netcdf_file(nav_group, "latitude", lat_info)

    chlor_a_info = {
        "data": np.ones((5, 5), np.float32),
        "dim_labels": ("number_of_lines", "pixels_per_line"),
        "attrs": {
            "long_name": "Chlorophyll Concentration, OCI Algorithm",
            "units": "mg m^-3",
            "standard_name": "mass_concentration_of_chlorophyll_in_sea_water",
            "valid_min": 0.001,
            "valid_max": 100.0,
        }
    }
    l2_flags = np.zeros((5, 5), dtype=np.int32)
    l2_flags[2, 2] = -1
    l2_flags_info = {
        "data": l2_flags,
        "dim_labels": ("number_of_lines", "pixels_per_line"),
        "attrs": {
            "valid_min": -2147483648,
            "valid_max": 2147483647,
        },
    }
    geophys_group = nc.createGroup("geophysical_data")
    _add_variable_to_netcdf_file(geophys_group, "chlor_a", chlor_a_info)
    _add_variable_to_netcdf_file(geophys_group, "l2_flags", l2_flags_info)
    nc.close()
    return [full_path]


def _add_variable_to_netcdf_file(nc, var_name, var_info):
    v = nc.createVariable(var_name, var_info["data"].dtype.str[1:], dimensions=var_info["dim_labels"],
                          fill_value=var_info.get("fill_value"))
    v[:] = var_info["data"]
    for attr_key, attr_val in var_info["attrs"].items():
        if isinstance(attr_val, (int, float)):
            attr_val = v.dtype.type(attr_val)
        setattr(v, attr_key, attr_val)



class TestSEADAS:
    """Test the SEADAS L2 file reader."""

    def test_available_reader(self):
        """Test that SEADAS L2 reader is available."""
        assert "seadas_l2" in available_readers()

    @pytest.mark.parametrize(
        "input_files",
        [
            lazy_fixture("seadas_l2_modis_chlor_a"),
            lazy_fixture("seadas_l2_viirs_npp_chlor_a"),
            lazy_fixture("seadas_l2_viirs_j01_chlor_a"),
        ])
    def test_scene_available_datasets(self, input_files):
        """Test that datasets are available."""
        scene = Scene(reader="seadas_l2", filenames=input_files)
        available_datasets = scene.all_dataset_names()
        assert len(available_datasets) > 0
        assert "chlor_a" in available_datasets

    @pytest.mark.parametrize(
        ("input_files", "exp_plat", "exp_sensor", "exp_rps"),
        [
            (lazy_fixture("seadas_l2_modis_chlor_a"), "Aqua", {"modis"}, 10),
            (lazy_fixture("seadas_l2_viirs_npp_chlor_a"), "Suomi-NPP", {"viirs"}, 16),
            (lazy_fixture("seadas_l2_viirs_j01_chlor_a"), "NOAA-20", {"viirs"}, 16),
            (lazy_fixture("seadas_l2_modis_chlor_a_netcdf"), "Terra", {"modis"}, 10),
        ])
    @pytest.mark.parametrize("apply_quality_flags", [False, True])
    def test_load_chlor_a(self, input_files, exp_plat, exp_sensor, exp_rps, apply_quality_flags):
        """Test that we can load 'chlor_a'."""
        reader_kwargs = {"apply_quality_flags": apply_quality_flags}
        scene = Scene(reader="seadas_l2", filenames=input_files, reader_kwargs=reader_kwargs)
        scene.load(["chlor_a"])
        data_arr = scene["chlor_a"]
        assert data_arr.dims == ("y", "x")
        assert data_arr.attrs["platform_name"] == exp_plat
        assert data_arr.attrs["sensor"] == exp_sensor
        assert data_arr.attrs["units"] == "mg m^-3"
        assert data_arr.dtype.type == np.float32
        assert isinstance(data_arr.attrs["area"], SwathDefinition)
        assert data_arr.attrs["rows_per_scan"] == exp_rps
        data = data_arr.data.compute()
        if apply_quality_flags:
            assert np.isnan(data[2, 2])
            assert np.count_nonzero(np.isnan(data)) == 1
        else:
            assert np.count_nonzero(np.isnan(data)) == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Benchmark SEVIRI HRIT operations."""

from __future__ import annotations

import os

from pyspectral.rayleigh import check_and_download as download_luts
from pyspectral.rsr_reader import check_and_download as download_rsr

from benchmarks.utils import GeoBenchmarks, get_filenames


class SEVIRIHRIT(GeoBenchmarks):
    """Benchmark SEVIRI HRIT reading."""

    timeout = 600
    data_files: list[str] = []
    subdir = os.path.join("seviri_hrit", "20180228_1500")
    reader = "seviri_l1b_hrit"

    def setup_cache(self):
        """Fetch the data files."""
        try:
            from satpy.demo import download_seviri_hrit_20180228_1500
            download_seviri_hrit_20180228_1500()
        except ImportError:
            assert len(get_filenames(self.subdir)) == 114  # nosec
        download_rsr()
        download_luts(aerosol_types=["rayleigh_only"])

    def setup(self):
        """Set up the benchmarks."""
        import dask.config
        self.data_files = get_filenames(self.subdir)
        dask.config.set({"array.chunk-size": "32MiB"})

    def time_load_one_channel(self):
        """Time the loading of one channel."""
        self.compute_channel("VIS006")

    def peakmem_load_one_channel(self):
        """Check peak memory usage of loading one channel."""
        self.compute_channel("VIS006")

    def time_load_overview(self):
        """Time the loading of the generation of overview."""
        self.compute_composite("overview")

    def peakmem_load_overview(self):
        """Check peak memory usage of the generation of overview."""
        self.compute_composite("overview")

    def time_save_overview_to_geotiff(self):
        """Time the generation and saving of overview."""
        self.save_composite_as_geotiff("overview")

    def peakmem_save_overview_to_geotiff(self):
        """Check peak memory usage of the generation and saving of overview."""
        self.save_composite_as_geotiff("overview")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unittesting the ASCAT SCATTEROMETER SOIL MOISTURE BUFR reader."""

import datetime as dt
import os
import sys
import unittest

import numpy as np

# TDB: this test is based on test_seviri_l2_bufr.py and test_iasi_l2.py

# This is a test for ASCAT SoilMoisture product message, take from a real
# bufr file distributed over EUMETCAST


def create_message():
    """Create fake message for testing."""
    nlat = 10
    nlon = 10
    samples = nlat*nlon
    lat, lon = np.meshgrid(np.linspace(63, 65, nlat), np.linspace(-30, -20, nlon))
    lat = np.round(np.ravel(lat), 4)
    lon = np.round(np.ravel(lon), 4)
    rstate = np.random.RandomState(0)
    surfaceSoilMoisture = np.round(rstate.rand(samples)*100, 1)
    surfaceSoilMoisture[0] = -1e+100
    retmsg = {
        "inputDelayedDescriptorReplicationFactor": [8],
        "edition": 4,
        "masterTableNumber": 0,
        "bufrHeaderCentre": 254,
        "bufrHeaderSubCentre": 0,
        "updateSequenceNumber": 0,
        "dataCategory": 12,
        "internationalDataSubCategory": 255,
        "dataSubCategory": 190,
        "masterTablesVersionNumber": 13,
        "localTablesVersionNumber": 0,
        "typicalYear": 2020,
        "typicalMonth": 12,
        "typicalDay": 21,
        "typicalHour": 9,
        "typicalMinute": 33,
        "typicalSecond": 0,
        "numberOfSubsets": samples,
        "observedData": 1,
        "compressedData": 1,
        "unexpandedDescriptors": 312061,
        "centre": 254,
        "subCentre": 0,
        "#1#softwareIdentification": 1000,
        "satelliteIdentifier": 4,
        "satelliteInstruments": 190,
        "year": 2020,
        "month": 12,
        "day": 21,
        "hour": 9,
        "minute": 33,
        "second": np.linspace(0, 59, samples),
        "latitude": lat,
        "longitude": lon,
        "surfaceSoilMoisture": surfaceSoilMoisture,
        "soilMoistureQuality": np.zeros(samples),
    }
    return retmsg


MSG = create_message()

# the notional filename that would contain the above test message data
FILENAME = "W_XX-EUMETSAT-TEST,SOUNDING+SATELLITE,METOPA+ASCAT_C_EUMC_20201221093300_73545_eps_o_125_ssm_l2.bin"
# the information that would be extracted from the above filename according to the pattern in the .yaml
FILENAME_INFO = {
    "reception_location": "TEST",
    "platform": "METOPA",
    "instrument": "ASCAT",
    "start_time": "20201221093300",
    "perigee": "73545",
    "species": "125_ssm",
    "level": "l2"
}

# file type info for the above file that is defined in the .yaml
FILETYPE_INFO = {
    "file_type": "ascat_l2_soilmoisture_bufr",
    "file_reader": "AscatSoilMoistureBufr"
}


def save_test_data(path):
    """Save the test file to the indicated directory."""
    import eccodes as ec
    filepath = os.path.join(path, FILENAME)
    with open(filepath, "wb") as f:
        for m in [MSG]:
            buf = ec.codes_bufr_new_from_samples("BUFR4_local_satellite")
            for key in m:
                val = m[key]
                if np.isscalar(val):
                    ec.codes_set(buf, key, val)
                else:
                    ec.codes_set_array(buf, key, val)
            ec.codes_set(buf, "pack", 1)
            ec.codes_write(buf, f)
            ec.codes_release(buf)
    return filepath


class TesitAscatL2SoilmoistureBufr(unittest.TestCase):
    """Test ASCAT Soil Mosture loader."""

    def setUp(self):
        """Create temporary file to perform tests with."""
        import tempfile

        from satpy.readers.ascat_l2_soilmoisture_bufr import AscatSoilMoistureBufr
        self.base_dir = tempfile.mkdtemp()
        self.fname = save_test_data(self.base_dir)
        self.fname_info = FILENAME_INFO
        self.ftype_info = FILETYPE_INFO
        self.reader = AscatSoilMoistureBufr(self.fname, self.fname_info, self.ftype_info)

    def tearDown(self):
        """Remove the temporary directory created for a test."""
        try:
            import shutil
            shutil.rmtree(self.base_dir, ignore_errors=True)
        except OSError:
            pass

    @unittest.skipIf(sys.platform.startswith("win"), "'eccodes' not supported on Windows")
    def test_scene(self):
        """Test scene creation."""
        from satpy import Scene
        fname = os.path.join(self.base_dir, FILENAME)
        scn = Scene(reader="ascat_l2_soilmoisture_bufr", filenames=[fname])
        assert "scatterometer" in scn.sensor_names
        assert dt.datetime(2020, 12, 21, 9, 33, 0) == scn.start_time
        assert dt.datetime(2020, 12, 21, 9, 33, 59) == scn.end_time

    @unittest.skipIf(sys.platform.startswith("win"), "'eccodes' not supported on Windows")
    def test_scene_load_available_datasets(self):
        """Test that all datasets are available."""
        from satpy import Scene
        fname = os.path.join(self.base_dir, FILENAME)
        scn = Scene(reader="ascat_l2_soilmoisture_bufr", filenames=[fname])
        assert "surface_soil_moisture" in scn.available_dataset_names()
        scn.load(scn.available_dataset_names())
        loaded = [dataset.name for dataset in scn]
        assert sorted(loaded) == sorted(scn.available_dataset_names())

    @unittest.skipIf(sys.platform.startswith("win"), "'eccodes' not supported on Windows")
    def test_scene_dataset_values(self):
        """Test loading data."""
        from satpy import Scene
        fname = os.path.join(self.base_dir, FILENAME)
        scn = Scene(reader="ascat_l2_soilmoisture_bufr", filenames=[fname])
        for name in scn.available_dataset_names():
            scn.load([name])
            loaded_values = scn[name].values
            fill_value = scn[name].attrs["fill_value"]
            # replace nans in data loaded from file with the fill value defined in the .yaml
            # to make them comparable
            loaded_values_nan_filled = np.nan_to_num(loaded_values, nan=fill_value)
            key = scn[name].attrs["key"]
            original_values = MSG[key]
            # this makes each assertion below a separate test from unittest's point of view
            # (note: if all subtests pass, they will count as one test)
            with self.subTest(msg="Test failed for dataset: "+name):
                assert np.allclose(original_values, loaded_values_nan_filled)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unittesting the EUMETSAT L2 BUFR reader."""

import sys
from datetime import datetime
from unittest import mock

import dask.array as da
import numpy as np
import pytest
from pyresample import geometry

from satpy.tests.utils import make_dataid

if sys.platform.startswith("win"):
    pytest.skip("'eccodes' not supported on Windows", allow_module_level=True)

AREA_DEF_MSG_IODC = geometry.AreaDefinition(
    "msg_seviri_iodc_48km",
    "MSG SEVIRI Indian Ocean Data Coverage service area definition with 48 km resolution",
    "",
    {"a": 6378169., "b": 6356583.8, "lon_0": 45.5,
     "h": 35785831., "proj": "geos", "units": "m"},
    232,
    232,
    (-5570248.686685662, -5567248.28340708, 5567248.28340708, 5570248.686685662)
)

AREA_DEF_MSG_FES = geometry.AreaDefinition(

    "msg_seviri_fes_48km",
    "MSG SEVIRI Full Earth Scanning service area definition with 48 km resolution",
    "",
    {"a": 6378169., "b": 6356583.8, "lon_0": 0.0,
     "h": 35785831., "proj": "geos", "units": "m"},
    232,
    232,
    (-5570248.6867, -5567248.2834, 5567248.2834, 5570248.6867)
)

AREA_DEF_MSG_IODC_EXT = geometry.AreaDefinition(
    "msg_seviri_iodc_9km_ext",
    "MSG SEVIRI Indian Ocean Data Coverage service area definition with 9 km resolution "
    "(extended outside original 3km grid)",
    "",
    {"a": 6378169., "b": 6356583.8, "lon_0": 45.5,
     "h": 35785831., "proj": "geos", "units": "m"},
    1238,
    1238,
    (-5571748.8883, -5571748.8882, 5571748.8882, 5571748.8883)
)

AREA_DEF_FCI_FES = geometry.AreaDefinition(
    "mtg_fci_fdss_32km",
    "MTG FCI Full Disk Scanning Service area definition with 32 km SSP resolution",
    "",
    {"x_0": 0, "y_0": 0, "ellps": "WGS84", "lon_0": 0.0,
     "h": 35786400., "proj": "geos", "units": "m"},
    348,
    348,
    (-5567999.998550739, -5567999.998550739, 5567999.994203017, 5567999.994203017)
)

AREA_DEF_MSG_FES_3km_ext = geometry.AreaDefinition(
    "msg_seviri_fes_9km_ext",
    "MSG SEVIRI Full Earth Scanning service area definition with 9 km resolution",
    "",
    {"a": 6378169., "b": 6356583.8, "lon_0": 0.0,
     "h": 35785831., "proj": "geos", "units": "m"},
    1238,
    1238,
    (-5571748.888268564, -5571748.888155806, 5571748.888155806, 5571748.888268564)
)

# Test data for mock file
DATA = np.random.default_rng().uniform(low=250, high=350, size=(128,))
LAT = np.random.default_rng().uniform(low=-80, high=80, size=(128,))
LON = np.random.default_rng().uniform(low=-38.5, high=121.5, size=(128,))

# Test cases dictionaries
TEST_DATA = {"GIIBUFRProduct_20231027140000Z_00_OMPEFS03_MET10_FES_E0000": {
                "platform_name": "MSG3",
                "spacecraft_number": "10",
                "RectificationLongitude": "E0000",
                "ssp_lon": 0.0,
                "area": AREA_DEF_MSG_FES_3km_ext,
                "seg_size": 3,
                "file_type": "seviri_l2_bufr_gii",
                "key": "#1#brightnessTemperature",
                "resolution": 9001,
                "data": DATA},
             "ASRBUFRProd_20231022224500Z_00_OMPEFS03_MET10_FES_E0000": {
                "platform_name": "MSG3",
                "spacecraft_number": "10",
                "RectificationLongitude": "E0000",
                "ssp_lon": 0.0,
                "area": AREA_DEF_MSG_FES,
                "seg_size": 16,
                "file_type": "seviri_l2_bufr_asr",
                "key": "#1#brightnessTemperature",
                "resolution": 48006,
                "data": DATA},
             "AMVBUFRProd_20231023044500Z_00_OMPEFS02_MET09_FES_E0455": {
                "platform_name": "MSG2",
                "spacecraft_number": "9",
                "RectificationLongitude": "E0455",
                "area": AREA_DEF_MSG_IODC,
                "ssp_lon": 45.5,
                "seg_size": None,
                "file_type": "seviri_l2_bufr_amv",
                "key": "#1#brightnessTemperature",
                "resolution": None,
                "data": DATA},
             "MSG2-SEVI-MSGASRE-0101-0101-20191106130000.000000000Z-20191106131702-1362128.bfr": {
                "platform_name": "MSG2",
                "spacecraft_number": "9",
                "RectificationLongitude": "E0455",
                "area": AREA_DEF_MSG_IODC,
                "ssp_lon": 45.5,
                "seg_size": 16,
                "file_type": "seviri_l2_bufr_asr",
                "key": "#1#brightnessTemperature",
                "resolution": 48006,
                "data": DATA},
             """W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-2-ASR--FD------BUFR_C_EUMT_
             20230623092246_L2PF_IV_20170410170000_20170410171000_V__C_0103_0000.bin""": {
                "platform_name": "MTGi1",
                "spacecraft_number": "24",
                "RectificationLongitude": "E0000",
                "area": AREA_DEF_FCI_FES,
                "ssp_lon": 0.0,
                "seg_size": 16,
                "file_type": "fci_l2_bufr_asr",
                "key": "#1#brightnessTemperature",
                "resolution": 32000,
                "data": DATA},
             """W_XX-EUMETSAT-Darmstadt,IMG+SAT,MTI1+FCI-2-AMV--FD------BUFR_C_EUMT_
20230623092246_L2PF_IV_20170410170000_20170410171000_V__C_0103_0000.bin""": {
                "platform_name": "MTGi1",
                "spacecraft_number": "24",
                "RectificationLongitude": "E0000",
                "area": AREA_DEF_FCI_FES,
                "ssp_lon": 0.0,
                "seg_size": None,
                "file_type": "fci_l2_bufr_amv",
                "key": "#1#brightnessTemperature",
                "resolution": None,
                "data": DATA}}

TEST_FILES = list(TEST_DATA.keys())


class L2BufrData:
    """Mock L2 BUFR data."""

    def __init__(self, filename, test_data, with_adef=False, rect_lon="default"):
        """Initialize by mocking test data for testing the L2 BUFR reader."""
        import eccodes as ec

        from satpy.readers.eum_l2_bufr import EumetsatL2BufrFileHandler

        self.buf1 = ec.codes_bufr_new_from_samples("BUFR4_local_satellite")
        ec.codes_set(self.buf1, "unpack", 1)
        # write the bufr test data twice as we want to read in and then concatenate the data in the reader
        # 55 id corresponds to METEOSAT 8
        ec.codes_set(self.buf1, "satelliteIdentifier", 47 + int(test_data["spacecraft_number"]))
        ec.codes_set_array(self.buf1, "#1#latitude", LAT)
        ec.codes_set_array(self.buf1, "#1#longitude", LON)
        ec.codes_set_array(self.buf1, test_data["key"], test_data["data"])

        self.m = mock.mock_open()

        FILETYPE_INFO = {"file_type": test_data["file_type"]}

        # only our offline product contain MPEF product headers so we get the metadata from there
        if ("BUFRProd" in filename):
            with mock.patch("satpy.readers.eum_l2_bufr.np.fromfile") as fromfile:

                MPEF_PRODUCT_HEADER = {
                    "NominalTime": datetime(2019, 11, 6, 18, 0),
                    "SpacecraftName": test_data["spacecraft_number"],
                    "RectificationLongitude": test_data["RectificationLongitude"]
                }
                fromfile.return_value = MPEF_PRODUCT_HEADER
                with mock.patch("satpy.readers.eum_l2_bufr.recarray2dict") as recarray2dict:
                    recarray2dict.side_effect = (lambda x: x)

                    FILENAME_INFO = {"start_time": "20231022224500", "spacecraft": test_data["platform_name"],
                                     "server": "TESTSERVER"}
                    self.fh = EumetsatL2BufrFileHandler(
                        filename, FILENAME_INFO, FILETYPE_INFO,
                        with_area_definition=with_adef,
                        rectification_longitude=int(test_data["RectificationLongitude"][1:]) / 10)
                    self.fh.mpef_header = MPEF_PRODUCT_HEADER

        else:
            # No Mpef Header  so we get the metadata from the BUFR messages
            with mock.patch("satpy.readers.eum_l2_bufr.open", self.m, create=True):
                with mock.patch("eccodes.codes_bufr_new_from_file",
                                side_effect=[self.buf1, None, self.buf1, None, self.buf1, None]) as ec1:
                    ec1.return_value = ec1.side_effect
                    with mock.patch("eccodes.codes_set") as ec2:
                        ec2.return_value = 1
                        with mock.patch("eccodes.codes_release") as ec5:
                            ec5.return_value = 1

                            FILENAME_INFO = {"start_time": "20191112000000",
                                             "spacecraft": test_data["platform_name"]}
                            self.fh = EumetsatL2BufrFileHandler(
                                filename, FILENAME_INFO, FILETYPE_INFO,
                                with_area_definition=with_adef,
                                rectification_longitude=int(test_data["RectificationLongitude"][1:]) / 10)

        self.resolution = test_data["resolution"]

    def get_data(self, dataset_name, key, coordinates):
        """Read data from mock file."""
        DATASET_INFO = {
            "name": dataset_name,
            "key": key,
            "fill_value": -1.e+100,
            "resolution": self.resolution
        }
        if coordinates:
            DATASET_INFO.update({"coordinates": ("longitude", "latitude")})

        with mock.patch("satpy.readers.eum_l2_bufr.open", self.m, create=True):
            with mock.patch("eccodes.codes_bufr_new_from_file",
                            side_effect=[self.buf1, self.buf1, None]) as ec1:
                ec1.return_value = ec1.side_effect
                with mock.patch("eccodes.codes_set") as ec2:
                    ec2.return_value = 1
                    with mock.patch("eccodes.codes_release") as ec5:
                        ec5.return_value = 1
                        z = self.fh.get_dataset(make_dataid(name=dataset_name, resolution=self.resolution),
                                                DATASET_INFO)

        return z


@pytest.mark.parametrize("input_file", TEST_FILES)
class TestL2BufrReader:
    """Test EUMETSAT L2 BUFR Reader."""

    @staticmethod
    def test_lonslats(input_file):
        """Test reading of longitude and latitude data with EUMETSAT L2 BUFR reader."""
        test_data = TEST_DATA[input_file]
        bufr_obj = L2BufrData(input_file, test_data)
        zlat = bufr_obj.get_data("latitude", "#1#latitude", coordinates=False)
        zlon = bufr_obj.get_data("longitude", "#1#longitude", coordinates=False)
        np.testing.assert_array_equal(zlat.values, np.concatenate((LAT, LAT), axis=0))
        np.testing.assert_array_equal(zlon.values, np.concatenate((LON, LON), axis=0))

    @staticmethod
    def test_attributes_with_swath_definition(input_file):
        """Test correctness of dataset attributes with data loaded with a SwathDefinition (default behaviour)."""
        test_data = TEST_DATA[input_file]
        bufr_obj = L2BufrData(input_file, test_data)
        z = bufr_obj.get_data(dataset_name="TestData", key=test_data["key"], coordinates=True)
        assert z.attrs["platform_name"] == test_data["platform_name"]
        assert z.attrs["ssp_lon"] == test_data["ssp_lon"]
        assert z.attrs["seg_size"] == test_data["seg_size"]

    @staticmethod
    def test_attributes_with_area_definition(input_file):
        """Test correctness of dataset attributes with data loaded with a AreaDefinition."""
        test_data = TEST_DATA[input_file]
        bufr_obj = L2BufrData(input_file, test_data, with_adef=True)
        _ = bufr_obj.get_data("latitude", "#1#latitude", coordinates=False)
        _ = bufr_obj.get_data("longitude", "#1#longitude", coordinates=False)

        z = bufr_obj.get_data(dataset_name="TestData", key=test_data["key"], coordinates=True)
        assert z.attrs["platform_name"] == test_data["platform_name"]
        assert z.attrs["ssp_lon"] == test_data["ssp_lon"]
        assert z.attrs["seg_size"] == test_data["seg_size"]

    @staticmethod
    def test_data_with_swath_definition(input_file):
        """Test data loaded with SwathDefinition (default behaviour)."""
        test_data = TEST_DATA[input_file]
        bufr_obj = L2BufrData(input_file, test_data)
        with pytest.raises(NotImplementedError):
            bufr_obj.fh.get_area_def(None)

        # concatenate original test arrays as get_dataset will have read and concatented the data
        x1 = np.concatenate((DATA, DATA), axis=0)
        z = bufr_obj.get_data(dataset_name="TestData", key=test_data["key"], coordinates=True)
        np.testing.assert_array_equal(z.values, x1)

    def test_data_with_area_definition(self, input_file):
        """Test data loaded with an area definition."""
        test_data = TEST_DATA[input_file]
        if test_data["seg_size"] is None:
            # Skip this test
            return

        bufr_obj = L2BufrData(input_file, test_data, with_adef=True)
        _ = bufr_obj.get_data("latitude", "#1#latitude", coordinates=False)
        _ = bufr_obj.get_data("longitude", "#1#longitude", coordinates=False)

        z = bufr_obj.get_data(dataset_name="TestData", key=test_data["key"], coordinates=True)

        ad = bufr_obj.fh.get_area_def(None)

        assert ad == test_data["area"]
        data_1d = np.concatenate((DATA, DATA), axis=0)

        # Put BUFR data on 2D grid that the 2D array returned by get_dataset should correspond to
        lons_1d, lats_1d = da.compute(bufr_obj.fh.longitude, bufr_obj.fh.latitude)
        icol, irow = ad.get_array_indices_from_lonlat(lons_1d, lats_1d)

        data_2d = np.empty(ad.shape)
        data_2d[:] = np.nan
        data_2d[irow.compressed(), icol.compressed()] = data_1d[~irow.mask]
        np.testing.assert_array_equal(z.values, data_2d)

        # Removed assert dedicated to products with seg_size=3 (covered by GII test case)

    def test_data_with_rect_lon(self, input_file):
        """Test data loaded with an area definition and a rectification longitude."""
        test_data = TEST_DATA[input_file]
        if test_data["seg_size"] is None:
            # Skip this test
            return

        bufr_obj = L2BufrData(input_file, test_data, with_adef=True)
        np.testing.assert_equal(bufr_obj.fh.ssp_lon,
                                int(test_data["RectificationLongitude"][1:]) / 10)
        _ = bufr_obj.get_data("latitude", "#1#latitude", coordinates=False)
        _ = bufr_obj.get_data("longitude", "#1#longitude", coordinates=False)
        _ = bufr_obj.get_data(dataset_name="TestData", key=test_data["key"], coordinates=True)
        # We need to lead the data in order to create the AreaDefinition

        ad = bufr_obj.fh.get_area_def(None)
        assert ad == test_data["area"]

    def test_resolution(self, input_file):
        """Test data loaded with the correct resolution attribute ."""
        test_data = TEST_DATA[input_file]
        bufr_obj = L2BufrData(input_file, test_data, with_adef=True)
        _ = bufr_obj.get_data("latitude", "#1#latitude", coordinates=False)
        _ = bufr_obj.get_data("longitude", "#1#longitude", coordinates=False)

        z = bufr_obj.get_data(dataset_name="TestData", key=test_data["key"], coordinates=True)

        assert z.attrs["resolution"] == bufr_obj.resolution

    def test_amv_with_area_def(self, input_file):
        """Test that AMV data can not be loaded with an area definition.

        The way to test this is to try load a variable with with_adef=True. The reader shall ignore this flag
        and return a 1D array, not a 2D.
        """
        test_data = TEST_DATA[input_file]
        if test_data["file_type"] not in ["seviri_l2_bufr_amv", "fci_l2_bufr_amv"]:
            return  # Skip this test for non AMV datasets

        bufr_obj = L2BufrData(input_file, test_data, with_adef=True)
        _ = bufr_obj.get_data("latitude", "#1#latitude", coordinates=False)
        _ = bufr_obj.get_data("longitude", "#1#longitude", coordinates=False)
        z = bufr_obj.get_data(dataset_name="TestData", key=test_data["key"], coordinates=True)

        assert len(z.dims) == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019.
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""A reader for files produced by the Hydrology SAF.

Currently this reader depends on the `pygrib` python package. The `eccodes`
package from ECMWF is preferred, but does not support python 3 at the time
of writing.

"""

import datetime as dt
import logging

import dask.array as da
import numpy as np
import pygrib
import xarray as xr
from pyresample import geometry

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

LOG = logging.getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()

CF_UNITS = {
    "none": "1",
}


class HSAFFileHandler(BaseFileHandler):
    """File handler for HSAF grib files."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the file handler."""
        super(HSAFFileHandler, self).__init__(filename,
                                              filename_info,
                                              filetype_info)

        self._msg_datasets = {}
        self._start_time = None
        self._end_time = None
        try:
            with pygrib.open(self.filename) as grib_file:
                first_msg = grib_file.message(1)
                analysis_time = self._get_datetime(first_msg)
                self._analysis_time = analysis_time
                self.metadata = self.get_metadata(first_msg)

        except (RuntimeError, KeyError):
            raise IOError("Unknown GRIB file format: {}".format(self.filename))

    @staticmethod
    def _get_datetime(msg):
        dtstr = str(msg["dataDate"]) + str(msg["dataTime"]).zfill(4)
        return dt.datetime.strptime(dtstr, "%Y%m%d%H%M")

    @property
    def analysis_time(self):
        """Get validity time of this file."""
        return self._analysis_time

    def get_metadata(self, msg):
        """Get the metadata."""
        try:
            center_description = msg["centreDescription"]
        except (RuntimeError, KeyError):
            center_description = None
        ds_info = {
            "filename": self.filename,
            "shortName": msg["shortName"],
            "long_name": msg["name"],
            "units": msg["units"],
            "centreDescription": center_description,
            "data_time": self._analysis_time,
            "nx": msg["Nx"],
            "ny": msg["Ny"],
            "projparams": msg.projparams
        }
        return ds_info

    def get_area_def(self, dsid):
        """Get area definition for message."""
        msg = self._get_message(1)
        try:
            return self._get_area_def(msg)
        except (RuntimeError, KeyError):
            raise RuntimeError("Unknown GRIB projection information")

    def _get_area_def(self, msg):
        """Get the area definition of the datasets in the file."""
        proj_param = msg.projparams.copy()

        Rx = 2 * np.arcsin(1. / msg["NrInRadiusOfEarth"]) / msg["dx"]
        Ry = 2 * np.arcsin(1. / msg["NrInRadiusOfEarth"]) / msg["dy"]

        x_0 = - msg["XpInGridLengths"]
        x_1 = msg["Nx"] - msg["XpInGridLengths"]
        y_0 = (msg["Ny"] - msg["YpInGridLengths"]) * -1
        y_1 = msg["YpInGridLengths"]

        min_x = (x_0 * Rx) * proj_param["h"]
        max_x = (x_1 * Rx) * proj_param["h"]

        min_y = (y_0 * Ry) * proj_param["h"]
        max_y = (y_1 * Ry) * proj_param["h"]

        area_extent = (min_x, min_y, max_x, max_y)

        area = geometry.AreaDefinition("hsaf_region",
                                       "A region from H-SAF",
                                       "geos",
                                       proj_param,
                                       msg["Nx"],
                                       msg["Ny"],
                                       area_extent)

        return area

    def _get_message(self, idx):
        with pygrib.open(self.filename) as grib_file:
            msg = grib_file.message(idx)
            return msg

    def get_dataset(self, ds_id, ds_info):
        """Read a GRIB message into an xarray DataArray."""
        if (ds_id["name"] not in self.filename):
            raise IOError("File does not contain {} data".format(ds_id["name"]))

        msg = self._get_message(1)

        ds_info = self.get_metadata(msg)
        ds_info["end_time"] = ds_info["data_time"]

        if (ds_id["name"] == "h05" or ds_id["name"] == "h05B"):
            flen = len(self.filename)
            timedelt = self.filename[flen-10:flen-8]
            ds_info["start_time"] = (ds_info["end_time"] -
                                     dt.timedelta(hours=int(timedelt)))
        else:
            ds_info["start_time"] = ds_info["end_time"]
        fill = msg["missingValue"]
        data = msg.values.astype(np.float32)
        if msg.valid_key("jScansPositively") and msg["jScansPositively"] == 1:
            data = data[::-1]

        if isinstance(data, np.ma.MaskedArray):
            data = data.filled(np.nan)
            data = da.from_array(data, chunks=CHUNK_SIZE)
        else:
            data[data == fill] = np.nan
            data = da.from_array(data, chunks=CHUNK_SIZE)

        return xr.DataArray(data, attrs=ds_info, dims=("y", "x"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""HRIT format reader.

References:
    ELECTRO-L GROUND SEGMENT MSU-GS INSTRUMENT,
      LRIT/HRIT Mission Specific Implementation, February 2012

"""

import datetime as dt
import logging

import numpy as np
import xarray as xr

from satpy.readers.core._geos_area import get_area_definition, get_area_extent
from satpy.readers.core.hrit import (
    HRITFileHandler,
    ancillary_text,
    annotation_header,
    base_hdr_map,
    image_data_function,
    time_cds_short,
)

logger = logging.getLogger("hrit_electrol")


# goms implementation:
key_header = np.dtype([("key_number", "u1"),
                       ("seed", ">f8")])

segment_identification = np.dtype([("GP_SC_ID", ">i2"),
                                   ("spectral_channel_id", ">i1"),
                                   ("segment_sequence_number", ">u2"),
                                   ("planned_start_segment_number", ">u2"),
                                   ("planned_end_segment_number", ">u2"),
                                   ("data_field_representation", ">i1")])

image_segment_line_quality = np.dtype([("line_number_in_grid", ">i4"),
                                       ("line_mean_acquisition",
                                        [("days", ">u2"),
                                         ("milliseconds", ">u4")]),
                                       ("line_validity", "u1"),
                                       ("line_radiometric_quality", "u1"),
                                       ("line_geometric_quality", "u1")])

goms_variable_length_headers = {
    image_segment_line_quality: "image_segment_line_quality"}

goms_text_headers = {image_data_function: "image_data_function",
                     annotation_header: "annotation_header",
                     ancillary_text: "ancillary_text"}

goms_hdr_map = base_hdr_map.copy()
goms_hdr_map.update({7: key_header,
                     128: segment_identification,
                     129: image_segment_line_quality
                     })


orbit_coef = np.dtype([("StartTime", time_cds_short),
                       ("EndTime", time_cds_short),
                       ("X", ">f8", (8, )),
                       ("Y", ">f8", (8, )),
                       ("Z", ">f8", (8, )),
                       ("VX", ">f8", (8, )),
                       ("VY", ">f8", (8, )),
                       ("VZ", ">f8", (8, ))])

attitude_coef = np.dtype([("StartTime", time_cds_short),
                          ("EndTime", time_cds_short),
                          ("XofSpinAxis", ">f8", (8, )),
                          ("YofSpinAxis", ">f8", (8, )),
                          ("ZofSpinAxis", ">f8", (8, ))])

cuc_time = np.dtype([("coarse", "u1", (4, )),
                     ("fine", "u1", (3, ))])

time_cds_expanded = np.dtype([("days", ">u2"),
                              ("milliseconds", ">u4"),
                              ("microseconds", ">u2"),
                              ("nanoseconds", ">u2")])


satellite_status = np.dtype([("TagType", "<u4"),
                             ("TagLength", "<u4"),
                             ("SatelliteID", "<u8"),
                             ("SatelliteName", "S256"),
                             ("NominalLongitude", "<f8"),
                             ("SatelliteCondition", "<u4"),
                             ("TimeOffset", "<f8")])

image_acquisition = np.dtype([("TagType", "<u4"),
                              ("TagLength", "<u4"),
                              ("Status", "<u4"),
                              ("StartDelay", "<i4"),
                              ("Cel", "<f8")])


prologue = np.dtype([("SatelliteStatus", satellite_status),
                     ("ImageAcquisition", image_acquisition, (10, )),
                     ("ImageCalibration", "<i4", (10, 1024))])


def recarray2dict(arr):
    """Change record array to a dictionary."""
    res = {}
    for dtuple in arr.dtype.descr:
        key = dtuple[0]
        ntype = dtuple[1]
        data = arr[key]
        if isinstance(ntype, list):
            res[key] = recarray2dict(data)
        else:
            res[key] = data

    return res


class HRITGOMSPrologueFileHandler(HRITFileHandler):
    """GOMS HRIT format reader."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the reader."""
        super(HRITGOMSPrologueFileHandler, self).__init__(filename, filename_info,
                                                          filetype_info,
                                                          (goms_hdr_map,
                                                           goms_variable_length_headers,
                                                           goms_text_headers))

        self.prologue = {}
        self.read_prologue()

    def read_prologue(self):
        """Read the prologue metadata."""
        with open(self.filename, "rb") as fp_:
            fp_.seek(self.mda["total_header_length"])
            data = np.fromfile(fp_, dtype=prologue, count=1)[0]

            self.prologue.update(recarray2dict(data))

        self.process_prologue()

    def process_prologue(self):
        """Reprocess prologue to correct types."""


radiometric_processing = np.dtype([("TagType", "<u4"),
                                   ("TagLength", "<u4"),
                                   ("RPSummary",
                                    [("Impulse", "<u4"),
                                     ("IsStrNoiseCorrection", "<u4"),
                                        ("IsOptic", "<u4"),
                                        ("IsBrightnessAligment", "<u4")]),
                                   ("OpticCorrection",
                                    [("Degree", "<i4"),
                                     ("A", "<f8", (16, ))]),
                                   ("RPQuality",
                                    [("EffDinRange", "<f8"),
                                     ("EathDarkening", "<f8"),
                                        ("Zone", "<f8"),
                                        ("Impulse", "<f8"),
                                        ("Group", "<f8"),
                                        ("DefectCount", "<u4"),
                                        ("DefectProcent", "<f8"),
                                        ("S_Noise_DT_Preflight", "<f8"),
                                        ("S_Noise_DT_Bort", "<f8"),
                                        ("S_Noise_DT_Video", "<f8"),
                                        ("S_Noise_DT_1_5", "<f8"),
                                        ("CalibrStability", "<f8"),
                                        ("TemnSKO", "<f8", (2, )),
                                        ("StructSKO", "<f8", (2, )),
                                        ("Struct_1_5", "<f8"),
                                        ("Zone_1_ 5", "<f8"),
                                        ("RadDif", "<f8")])])

geometric_processing = np.dtype([("TagType", "<u4"),
                                 ("TagLength", "<u4"),
                                 ("TGeomNormInfo",
                                  [("IsExist", "<u4"),
                                   ("IsNorm", "<u4"),
                                      ("SubLon", "<f8"),
                                      ("TypeProjection", "<u4"),
                                      ("PixInfo", "<f8", (4, ))]),
                                 ("SatInfo",
                                  [("TISO",
                                    [("T0", "<f8"),
                                     ("dT", "<f8"),
                                        ("ASb", "<f8"),
                                        ("Evsk", "<f8", (3, 3, 4)),
                                        ("ARx", "<f8", (4, )),
                                        ("ARy", "<f8", (4, )),
                                        ("ARz", "<f8", (4, )),
                                        ("AVx", "<f8", (4, )),
                                        ("AVy", "<f8", (4, )),
                                        ("AVz", "<f8", (4, ))]),
                                      ("Type", "<i4")]),
                                 ("TimeProcessing", "<f8"),
                                 ("ApriorAccuracy", "<f8"),
                                 ("RelativeAccuracy", "<f8", (2, ))])

epilogue = np.dtype([("RadiometricProcessing", radiometric_processing, (10, )),
                     ("GeometricProcessing", geometric_processing, (10, ))])
# FIXME: Add rest of the epilogue


class HRITGOMSEpilogueFileHandler(HRITFileHandler):
    """GOMS HRIT format reader."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the reader."""
        super(HRITGOMSEpilogueFileHandler, self).__init__(filename, filename_info,
                                                          filetype_info,
                                                          (goms_hdr_map,
                                                           goms_variable_length_headers,
                                                           goms_text_headers))
        self.epilogue = {}
        self.read_epilogue()

    def read_epilogue(self):
        """Read the prologue metadata."""
        with open(self.filename, "rb") as fp_:
            fp_.seek(self.mda["total_header_length"])
            data = np.fromfile(fp_, dtype=epilogue, count=1)[0]

            self.epilogue.update(recarray2dict(data))


C1 = 1.19104273e-5
C2 = 1.43877523


# Defined in MSG Level 1.5 Image Data Format Description
# https://user.eumetsat.int/s3/eup-strapi-media/pdf_ten_05105_msg_img_data_e7c8b315e6.pdf
SPACECRAFTS = {19001: "Electro-L N1",
               19002: "Electro-L N2",
               19003: "Electro-L N3"}


class HRITGOMSFileHandler(HRITFileHandler):
    """GOMS HRIT format reader."""

    def __init__(self, filename, filename_info, filetype_info,
                 prologue, epilogue):
        """Initialize the reader."""
        super(HRITGOMSFileHandler, self).__init__(filename, filename_info,
                                                  filetype_info,
                                                  (goms_hdr_map,
                                                   goms_variable_length_headers,
                                                   goms_text_headers))
        self.prologue = prologue.prologue
        self.epilogue = epilogue.epilogue
        self.chid = self.mda["spectral_channel_id"]
        sublon = self.epilogue["GeometricProcessing"]["TGeomNormInfo"]["SubLon"]
        sublon = sublon[self.chid]
        self.mda["projection_parameters"]["SSP_longitude"] = np.rad2deg(sublon)
        self.mda["orbital_parameters"]["satellite_nominal_longitude"] = np.rad2deg(
            self.prologue["SatelliteStatus"]["NominalLongitude"])
        satellite_id = self.prologue["SatelliteStatus"]["SatelliteID"]
        self.platform_name = SPACECRAFTS[satellite_id]

    def get_dataset(self, key, info):
        """Get the data  from the files."""
        res = super(HRITGOMSFileHandler, self).get_dataset(key, info)

        res = self.calibrate(res, key["calibration"])
        res.attrs["units"] = info["units"]
        res.attrs["standard_name"] = info["standard_name"]
        res.attrs["wavelength"] = info["wavelength"]
        res.attrs["platform_name"] = self.platform_name
        res.attrs["sensor"] = "msu-gs"
        res.attrs["orbital_parameters"] = {
            "satellite_nominal_longitude": self.mda["orbital_parameters"]["satellite_nominal_longitude"],
            "satellite_nominal_latitude": 0.,
            "projection_longitude": self.mda["projection_parameters"]["SSP_longitude"],
            "projection_latitude": 0.,
            "projection_altitude": 35785831.00
        }

        return res

    def calibrate(self, data, calibration):
        """Calibrate the data."""
        tic = dt.datetime.now()
        if calibration == "counts":
            res = data
        elif calibration in ["radiance", "brightness_temperature"]:
            res = self._calibrate(data)
        else:
            raise NotImplementedError("Don't know how to calibrate to " +
                                      str(calibration))

        res.attrs["standard_name"] = calibration
        res.attrs["calibration"] = calibration

        logger.debug("Calibration time " + str(dt.datetime.now() - tic))
        return res

    @staticmethod
    def _getitem(block, lut):
        return lut[block]

    def _calibrate(self, data):
        """Visible/IR channel calibration."""
        lut = self.prologue["ImageCalibration"][self.chid]
        if abs(lut).max() > 16777216:
            lut = lut.astype(np.float64)
        else:
            lut = lut.astype(np.float32)
        lut /= 1000
        lut[0] = np.nan
        # Dask/XArray don't support indexing in 2D (yet).
        res = data.data.map_blocks(self._getitem, lut, dtype=lut.dtype)
        res = xr.DataArray(res, dims=data.dims,
                           attrs=data.attrs, coords=data.coords)
        res = res.where(data > 0)
        return res

    def get_area_def(self, dsid):
        """Get the area definition of the band."""
        pdict = {}
        pdict["cfac"] = np.int32(self.mda["cfac"])
        pdict["lfac"] = np.int32(self.mda["lfac"])
        pdict["coff"] = np.float32(self.mda["coff"])
        pdict["loff"] = np.float32(self.mda["loff"])

        pdict["a"] = 6378169.00
        pdict["b"] = 6356583.80
        pdict["h"] = 35785831.00
        pdict["scandir"] = "N2S"

        pdict["ssp_lon"] = self.mda["projection_parameters"]["SSP_longitude"]

        pdict["nlines"] = int(self.mda["number_of_lines"])
        pdict["ncols"] = int(self.mda["number_of_columns"])

        pdict["loff"] = pdict["nlines"] - pdict["loff"]

        pdict["a_name"] = "geosgoms"
        pdict["a_desc"] = "Electro-L/GOMS channel area"
        pdict["p_id"] = "goms"

        area_extent = get_area_extent(pdict)
        area = get_area_definition(pdict, area_extent)

        self.area = area

        return area
